﻿using System;
using System.Collections.Generic;
using System.Diagnostics.Eventing.Reader;
using System.Linq;
using System.Runtime.CompilerServices;
using System.Runtime.Remoting.Messaging;
using System.Text;
using System.Threading.Tasks;
using HeuristicLab.Common;

namespace HeuristicLab.Problems.GrammaticalOptimization {
  public class ExpressionInterpreter {
    private static readonly double[] emptyErc = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };

    private string sentence;
    private int syIdx;
    // interprets sentences from L(G(Expr)):
    // Expr -> Term { ('+' | '-' | '^' ) Term } 
    // Term -> Fact { ('*' | '%') Fact }
    // Fact -> '!' Expr | '(' Expr ')' | Var | const
    // Var -> 'a'..'z'
    // const -> '0' .. '9'

    // uses protected division symbol %
    // constants are Koza-style ephemeral random constants (ERC). for now we only allow up to 10 constants.
    // The constant symbols '0' .. '9' are treated as ERC indices

    // for boolean problems the symbol * representes the AND operator, the symbol + represents the OR operator, ! = NOT, ^ = XOR

    public bool Interpret(string sentence, bool[] vars) {
      return Interpret(sentence, vars, emptyErc);
    }

    public bool Interpret(string sentence, bool[] vars, double[] erc) {
      InitLex(sentence);
      var d = new double[vars.Length];
      for (int i = 0; i < vars.Length; i++) d[i] = vars[i] ? 1.0 : 0.0;
      return !DoubleExtensions.IsAlmost(Expr(d, erc), 0);
    }

    public double Interpret(string sentence, double[] vars) {
      return Interpret(sentence, vars, emptyErc);
    }

    public double Interpret(string sentence, double[] vars, double[] erc) {
      InitLex(sentence);
      return Expr(vars, erc);
    }


    private void InitLex(string sentence) {
      this.sentence = sentence;
      this.syIdx = 0;
    }

    private char CurSy() {
      if (syIdx >= sentence.Length) return '\0';
      return sentence[syIdx];
    }
    private void NewSy() {
      if (syIdx < sentence.Length) syIdx++;
    }

    // helper for xor
    private double Not(double x) {
      return DoubleExtensions.IsAlmost(x, 0) ? 1.0 : 0.0;
    }

    private double Expr(double[] d, double[] erc) {
      var r = 0.0;
      r = Term(d, erc);
      var curSy = CurSy();
      while (curSy == '+' || curSy == '-' || curSy == '^') {
        if (curSy == '+') {
          NewSy();
          r += Term(d, erc);
        } else if (curSy == '-') {
          NewSy();
          r -= Term(d, erc);
        } else {
          NewSy();
          var e = Expr(d, erc);
          r = Not(r) * e + r * Not(e); // xor = (!x AND y) OR (x AND !y)
        }
        curSy = CurSy();
      }
      return r;
    }

    private double Term(double[] d, double[] erc) {
      var r = 0.0;
      r = Fact(d, erc);
      var curSy = CurSy();
      while (curSy == '*' || curSy == '%') {
        if (curSy == '*') {
          NewSy();
          r *= Fact(d, erc);
        } else {
          NewSy();
          var nom = Fact(d, erc);
          if (HeuristicLab.Common.Extensions.IsAlmost(nom, 0.0)) nom = 1.0;
          r /= nom;
        }
        curSy = CurSy();
      }
      return r;
    }

    private double Fact(double[] d, double[] erc) {
      double r = 0.0;
      var curSy = CurSy();
      if (curSy == '!') {
        NewSy();
        r = Not(Expr(d, erc));
      } else if (curSy == '(') {
        NewSy();
        r = Expr(d, erc);
        if (CurSy() != ')') throw new ArgumentException();
        NewSy();
      } else if (curSy >= 'a' && curSy <= 'z') {
        int o = (byte)curSy - (byte)'a';
        //int o = Convert.ToByte(CurSy()) - Convert.ToByte('a');
        if (o < 0 || o >= d.Length) throw new ArgumentException();
        r = d[o];
        NewSy();
      } else if (curSy == '/') {
        // /-symbol is used in the expressionextender to represent inverse (1/x). 
        // this is necessary because we also use symbols 0..9 as indices for ERCs
        NewSy();
        r = 1.0 / Fact(d, erc);
      } else /* if (curSy >= '0' && curSy <= '9') */ {
        int o = (byte)curSy - (byte)'0';
        //int o = Convert.ToByte(CurSy()) - Convert.ToByte('a');
        if (o < 0 || o >= 10) throw new ArgumentException();
        r = erc[o];
        NewSy();
      }
      //} else throw new ArgumentException();
      return r;
    }

  }
}
